`default_nettype none
module decimation(hpsel, mod, mck, wck, lrck, bck, i2sdata, fclipled);
input wire [3:0] hpsel; // select high-pass frequency for second-order IIR high-pass
// Pull-ups on all hpsel wires, coding:
// 4'b1110: 0 Hz except during the first few seconds (autocalibration mode)
// 4'b1101: 1.2 Hz at 88.2 kHz sample rate
// 4'b1011: 4.8 Hz
// 4'b0111: 9.7 Hz
// default: 1.2 Hz
input wire signed [6:1] mod; // Rumour has it that it's offset binary. The strange bit
// numbering comes straight from Texas Instruments.
reg signed [5:0] din; // two's complement, normal bit numbering, clocked with mck
input wire mck; // modulator master clock, 11.2896 MHz or 12.288 MHz
input wire wck; // modulator word clock, master clock frequency divided by 2
reg wckc1, wckc; // wck clocked in on the falling mck edge and then on the rising edge
reg signed [5:0] dinf;
reg wckfold, wckf, chan_in1;
wire ndl, ndr;

output reg lrck; // I2S word clock, a.k.a. left-right clock
output reg bck; // I2S bit clock
output reg i2sdata; // I2S data line

output wire fclipled; // can be used to drive a clipping LED that will light up 
// for about a second when the output signal is too large to fit in 24 bits,
// which can only happen with very large and unusually-shaped signals

(* ASYNC_REG = "TRUE" *) reg [3:0] hpsel1;
(* ASYNC_REG = "TRUE" *) reg [3:0] hpsel2;
(* ASYNC_REG = "TRUE" *) reg [3:0] hpsel3;

reg signed [64:0] dither;
wire signed [63:0] random;
reg signed [63:0] prevrandom;
wire CLK0, CLK180, CLK270, CLK2X, CLK2X180, CLK90, CLKDV, LOCKED, PSDONE, STATUS, CLK0;
wire firclock, firclock180;
reg locked_long, locked_synch1, locked_synch2;
wire endofstartup, reset0, reset1, reset2, reset3, resetmck;
reg [19:0] delaycount20;
reg firreset;

wire signed [29:0] dout1;
reg signed [29:0] din2;
wire signed [62:0] dout2;
wire signed [26:0] dout3;
reg signed [23:0] dout4;
wire rdy1, rdy2, rdy3;
reg rdy4;

wire nd1;
reg nd2;

wire rfd1, chan_sync1, chan_out1, rfd2, chan_in2, chan_out2;

reg [1:0] freqsel;

reg [30:0] firstfiveseconds;

reg fclip;
reg [28:0] fclipcounter;

reg [63:0] piso;
// reg [63:0] pisolr;
reg [5:0] cntlr;
reg signed [23:0] left;
reg signed [23:0] leftt;
reg signed [23:0] right;
reg lrrdy, ack;

reg adcclip;

PULLUP hps0(hpsel[0]);
PULLUP hps1(hpsel[1]);
PULLUP hps2(hpsel[2]);
PULLUP hps3(hpsel[3]);

// always @ (posedge mck) din[5:0] <= {!mod[6], mod[5:1]}; // Correct assignment if mod[6:1] is indeed in offset binary notation
always @ (posedge mck) din[5:0] <= mod[6:1]; // Correct assignment if mod[6:1] is indeed in two's complement notation
always @ (negedge mck) wckc1 <= wck;
always @ (posedge mck) wckc <= wckc1;
// when wckc becomes 1, the left channel data are clocked in


// Generating the power-on reset

STARTUP_SPARTAN6 startding(.EOS(endofstartup), .CLK(1'b0), .GSR(1'b0), .GTS(1'b0), .KEYCLEARB(1'b1));

assign reset0 = !endofstartup;
(* ASYNC_REG = "TRUE" *) FDPE flipflop1(reset1, mck, 1'b1, reset0, reset0);
(* ASYNC_REG = "TRUE" *) FDPE flipflop2(reset2, mck, 1'b1, reset1, reset1);
(* ASYNC_REG = "TRUE" *) FDPE flipflop3(reset3, mck, 1'b1, reset2, reset2);
// Make some extra delay to give the mck crystal oscillator a chance
// to settle properly, no idea if that is really needed.
//
// I don't see any way to guarantee that nothing ends up in a wrong state when the
// master clock is switched between 11.2896 MHz and 12.288 MHz or the other way around,
// but one can try to minimize the probability by resetting as much as
// possible when the DLL loses lock (that is, use the firreset that is generated a bit
// further down for all major blocks).
// 
always @(posedge mck or posedge reset3)
begin
	if (reset3)
		delaycount20 <= 20'b0;
	else
		if (!delaycount20[19]) delaycount20 <= delaycount20 + 20'b1;
end
// FDPE bla (output, clock, clock enable, asynchronous preset, data)
FDPE flipflop7b (resetmck, mck, 1'b1, reset3, !delaycount20[19]);

// Making an internal clock signal
// firclock: 196.608 MHz at 12.288 MHz master clock frequency, corresponding
// to 96 kHz output sample rate, 180.6336 MHz 11.2896 MHz master clock
// frequency, corresponding to 88.2 kHz output sample rate.
//
// Clock multiplication to be done by a DLL and not a PLL,
// because Spartan 6 PLLs can't work with reference frequencies
// below 19 MHz.
DCM_SP #(.CLKDV_DIVIDE(2.0), //Divided output related to input frequency, not used
	.CLKFX_DIVIDE(1),
	.CLKFX_MULTIPLY(16), // Produces a 16 times multiplied clock frequency, 16 times the 11.2896 MHz or 12.288 MHz master clock
	.CLKIN_DIVIDE_BY_2("FALSE"),
	.CLKIN_PERIOD(88.577),
	.CLKOUT_PHASE_SHIFT("FIXED"),
	.CLK_FEEDBACK("1X"),
	.DESKEW_ADJUST("SOURCE_SYNCHRONOUS"), // ("SYSTEM_SYNCHRONOUS"),
	.DFS_FREQUENCY_MODE("LOW"),
	.DLL_FREQUENCY_MODE("LOW"),
	.DSS_MODE("NONE"),
	.DUTY_CYCLE_CORRECTION("TRUE"),
	.FACTORY_JF(16'hc080),
	.PHASE_SHIFT(0), // 0/256 times input clock period
	.STARTUP_WAIT("FALSE"))
	DCM_SP_inst (CLK0, CLK180, CLK270, CLK2X, CLK2X180, CLK90, CLKDV, firclock, firclock180, LOCKED, PSDONE, STATUS, CLK0, mck, 1'b0, 1'b0, 1'b0, 1'b0, resetmck);
	// Connecting CLKFB to CLK0 is actually not even needed when you only use CLKFX.
	// Use a reset that is only released after mck is properly settled!

always @ (posedge firclock or negedge LOCKED)
begin
	if (~LOCKED)
	begin
		locked_long <= 1'b0;
		locked_synch1 <= 1'b0;
		locked_synch2 <= 1'b0;
	end
	else
	begin
		locked_synch1 <= LOCKED;
		locked_synch2 <= locked_synch1;
		locked_long <= locked_synch2;
	end
end

always @ (posedge firclock or posedge resetmck)
begin
	if (resetmck)
		firreset <= 1'b1;
	else
		firreset <= (! locked_long);
end
// When the DLL loses lock, for example because the master clock is switched, 
// locked_long will be low until at least 3 firclock cycles after LOCKED goes
// high again. Hence, there will be a firreset pulse, but no resetmck pulse.
// As the offset may change when switching the master clock, the
// autocalibration needs to be redone and firstfiveseconds can best work on 
// the firclock and firreset.

// Make nd1 pulses. The channel also has to be correct (left-right).

always @ (posedge firclock)
begin
	dinf <= din;
	wckfold <= wckf;
	wckf <= wckc;
end

assign ndl = wckf && (!wckfold);
assign ndr = (!wckf) && wckfold;
assign nd1 = ((ndl && (!chan_in1)) || (ndr && chan_in1)) && rfd1;
// channel 0: left, channel 1: right
// Unfortunately the CIC filter has no chan_in output

always @ (posedge firclock)
begin
	if (firreset)
		chan_in1 <= 1'b0;
	else
		chan_in1 <= chan_in1 ^^ nd1;
end	

CIC16 CIC16LR (
	.din(dinf), // input [5 : 0] din
	.nd(nd1), // input nd
	.sclr(firreset), // input sclr
	.clk(firclock), // input clk
	.dout(dout1), // output [29 : 0] dout
	.rdy(rdy1), // output rdy
	.rfd(rfd1), // output rfd
	.chan_sync(chan_sync1), // output chan_sync
	.chan_out(chan_out1)); // output [0 : 0] chan_out



// rdy van de eerste aan nd van de tweede?
//
// Wat met chan_out?
// chan_in van decim4 is ook een uitgang.
// Block interface: RDY blijft zoveel klokpulsen hoog als er kanalen zijn
// en de data van de verschillende kanalen komen er dan achter elkaar uit.
// Streaming interface: elk kanaal zijn eigen RDY-puls van 1 klokpuls breed.
// Streaming past het beste op de ingang van het FIR-filter.

always @ (posedge firclock)
begin
	din2 <= dout1;
	nd2 <= rdy1 && (!(chan_out1 ^^ chan_in2));
end

decim4 decim4LR (
	.sclr(firreset), // input sclr
	.clk(firclock), // input clk
	.nd(nd2), // input nd
	.rfd(rfd2), // output rfd
	.rdy(rdy2), // output rdy
	.chan_in(chan_in2), // output [0 : 0] chan_in
	.chan_out(chan_out2), // output [0 : 0] chan_out
	.din(din2), // input [29 : 0] din
	.dout(dout2)); // output [62 : 0] dout

// Clipping detector for the ADC. TI claims it can be driven to 90 %,
// see https://e2e.ti.com/support/audio-group/audio/f/audio-forum/132942/pcm4222-modulator-output,
// although the measurements from Charles in that thread even go beyond 90 % (which he fails to understand). 
// An offset binary code of 111111 or 000000 at the modulator output produces
// 001111100000... or 110000000... at dout2. Hence, 90 % is about
// 0011100000... or 1100011111...
always @ (posedge firclock)
begin
	if (rdy2) adcclip <= ((!dout2[62]) && (dout2[61] || (dout2[60] && dout2[59] && dout2[58]))) || (dout2[62] && ((!dout2[61]) || ((!dout2[60]) && (!dout2[59]) && (!dout2[58]))));
end

// Selecting the cut-off frequency of the IIR high-pass filter

// Synchronize the asynchronous input from the hpsel switch, no need for debouncing
always @ (posedge mck)
begin
	hpsel1 <= hpsel;
	hpsel2 <= hpsel1;
end

// Choose the correct freqsel value: normally this just follows the
// synchronized hpsel value, but for the autocalibration mode, it also depends
// on the time since power-on or since the last time the DLL has lost lock
always @ (posedge firclock)
begin
	if (firreset)
		firstfiveseconds <= 31'b0;
	else
		if (!firstfiveseconds[30]) firstfiveseconds <= firstfiveseconds + 31'b1;
	hpsel3 <= hpsel2;
	case (hpsel3)
		4'b1110: 					// When the autocalibration mode is chosen:
			begin
				if (!firstfiveseconds[30])  	// if first few seconds haven't passed yet
					freqsel <= 2'b01;   	// set filter to 1.2 Hz bandwidth
				else
					freqsel <= 2'b00;   	// else freeze filter
			end
		4'b1101: freqsel <= 2'b01;
	       	4'b1011: freqsel <= 2'b10;
		4'b0111: freqsel <= 2'b11;
		default: freqsel <= 2'b01;
	endcase
end

// Rounding and high-pass filtering
iirhp iirhp1(dout3, rdy3, dout2, rdy2, dither, freqsel, firclock, firreset);

// Going from 27 bits to 24 bits by discarding the highest bits and applying
// a clipper; the highest four bits will normally be equal, unless a very
// unusual waveform causes a large filter overshoot
always @ (posedge firclock)
begin
	case (dout3[26:23])
		4'b1111:
		begin
			fclip <= 0;
			dout4[23:0] <= dout3[23:0];
		end
		4'b0000:
		begin
			fclip <= 0;
			dout4[23:0] <= dout3[23:0];
		end
		default:
		begin
			fclip <= 1;
			if (dout3[26])
				dout4[23:0] <= 24'sb100000000000000000000000;
			else
				dout4[23:0] <= 24'sb011111111111111111111111;
		end
	endcase
	rdy4 <= rdy3;
end

// Stretching the fclip or adcclip pulse so you can drive a clipping LED with it
always @ (posedge firclock)
begin
	if (firreset)
		fclipcounter <= 29'b10000000000000000000000000000;
	else
	begin
		if (fclip || adcclip)
			fclipcounter <= 29'b0;
		else
			if (!fclipcounter[28]) fclipcounter <= fclipcounter + 29'b1;
	end
end

assign fclipled = !fclipcounter[28];

// Dither generation for the rounding stage inside the IIR filter,
// the seed is some arbitrarily chosen 64-bit number
xorshiftplus xorshiftplus1(firclock, firreset, 64'h1AB3CD049559ABCD, random);
always @ (posedge firclock)
begin
	prevrandom <= random;
	dither <= random + prevrandom;
end

// I2S interface
always @ (posedge firclock)
begin
	if (rdy4 && (!chan_out2)) left <= dout4;
	if (rdy4 && chan_out2)
	begin
		right <= dout4;
		leftt <= left;
	end
	lrrdy <= ((rdy4 && chan_out2) || (lrrdy && (!ack))) && (!firreset);
end
always @ (posedge mck)
begin
	if (resetmck)
		bck <= 1'b0;
	else
		bck <= !bck;
	if (bck)
	begin
		i2sdata <= piso[63];
		// lrck <= pisolr[63];
		lrck <= cntlr[5];
		if (lrrdy)
		begin
			piso[63:0] <= {1'b0, leftt, 8'b0, right, 7'b0};
			cntlr <= 6'b000000;
			// pisolr[63:0] <= 64'b0000000000000000000000000000000011111111111111111111111111111111;
			ack <= 1'b1;
		end
		else
		begin
			piso[63:0] <= {piso[62:0], 1'b0};
			// pisolr[63:0] <= {pisolr[62:0], pisolr[63]};
			cntlr <= (cntlr + 6'b000001) % 64;
			ack <= 1'b0;
		end
	end
end

endmodule
`default_nettype wire
